/*
 * Copyright (c) 1982, 1986, 1988, 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)uipc_mbuf.c	8.2 (Berkeley) 1/4/94
 * $FreeBSD: src/sys/kern/uipc_mbuf.c,v 1.51.2.37 2004/06/17 00:08:22 fjoe Exp $
 */


#include <sys/types.h>
#include <sys/param.h>
#include <sys/mbuf.h>
#include <net/if_var.h>
#include <kern/mm.h>
#include <kern/buf.h>
#include <kern/lib.h>
#include <kern/test.h>


//================================== PRIVATE ============================================


enum{
	MBUF_MAX = 0x10000,		// f[^TCY
};

static ALLOC_BUF AllocMbuf;


//================================== PUBLIC =============================================


// mbuf蓖Ă
// return : 蓖Ămbuf or NULLiG[j
struct mbuf *m_allocm(const uint size)
{
	struct mbuf *mbuf;
	int rest;
	
	if (size == 0){
		return NULL;
	}
	
	mbuf = kmalloc(sizeof(*mbuf));
	if (mbuf == NULL){
		return NULL;
	}
	memset(mbuf, 0, sizeof(*mbuf));
	rest = allocBuf(size, &AllocMbuf, (void**)&mbuf->m_data);
	if (rest == -EFAULT){
		printk("m_allocm() error! : total max size over.\n");
		return NULL;
	}
	if (rest != NOERR){
		return NULL;
	}
	mbuf->mh_len = size;
	
	return mbuf;
}


// mbufJ
void m_freem(struct mbuf *mbuf)
{
	if (mbuf != NULL){
		// Oobt@͊Js
		if ((mbuf->m_flags & M_EXT) == 0){
			freeBuf(mbuf->m_data, &AllocMbuf);
		}
		kfree(mbuf);
//**********************************************
		if (test_mmFlag == 1){
			test_mm(*((uint*)&mbuf - 1));
		}
//**********************************************
	}
}


// Rs[AhX
// return : mbuf or NULL
struct mbuf *m_devget(
	char *fromBuf,				// Rs[
	int totlen, 				// Rs[TCY
	int off,
	struct ifnet *ifp,
	void (*copy)(char *from, caddr_t to, u_int len))
{
	struct mbuf *mbuf = ifp->mRecBuf;
	ASSERT(mbuf != NULL);
	memcpy(mbuf->m_data + off, fromBuf, totlen);
	mbuf->mh_len = totlen;
	mbuf->mh_off = off;

	return mbuf;
}


// obt@|C^̈ړ
void m_adj(
	struct mbuf *mbuf,		// Ώۃobt@
	int off)				// |C^ړItZbg
{
	mbuf->mh_off += off;
}


/*
 * Copy data from a buffer back into the indicated mbuf chain,
 * starting "off" bytes from the beginning, extending the mbuf
 * chain if necessary.
 */
void m_copyback(
	struct	mbuf *mbuf,		// Rs[
	int off,				// Rs[ItZbg
	int len,				// Rs[TCY
	caddr_t cp)				// Rs[
{
	ASSERT(off + len <= mbuf->mh_len);
	memcpy(mbuf->m_data + mbuf->mh_off + off, cp, len);
}


void initMbuf()
{
	initAllocBuf(MBUF_MAX, &AllocMbuf);
}


// mbuf\̂̂݊蓖Ă
//void mgethdr(struct mbuf **o_mbuf)
//{
//	*o_mbuf = kmalloc(sizeof(*o_mbuf));
//}


/*
 * Copy data from an mbuf chain starting "off" bytes from the beginning,
 * continuing for "len" bytes, into the indicated buffer.
 */
//void m_copydata(
//	struct mbuf *mbuf,		// Rs[
//	int off,				// Rs[obt@ItZbg
//	int len,				// Rs[TCY
//	caddr_t cp)				// Rs[
//{
//	memcpy(cp, mbuf->m_data + off, len);
//	mbuf->m_data = NULL;
//}
